import sys, datetime, mox
from django.core import mail
from django.test import TestCase
from django.test.client import Client
from urlimaging.models import *
from urlimaging.validator import *
import boto.s3.key


GOOD_IMAGE = 'patrickomatic.com/photography/brewing.jpg'

class CommandRunnerTest(TestCase):
	fixtures = ['image']

	def setUp(self):
		self.cr = CommandRunner()
		self.mox = mox.Mox()
		self.site = Site.objects.get(domain_name='patrickomatic.com')

	def tearDown(self):
		self.mox.UnsetStubs()


	def test_parse_url__resize(self):
		self.assert_good_urls([
				'resize/50x50/' + GOOD_IMAGE,
				'resize/1x1/www.google.com/Logos/Brewing.jpg',
		])

	def test_parse_url__scale(self):
		self.assert_good_urls([
				'scale/1/' + GOOD_IMAGE,
				'scale/50/' + GOOD_IMAGE,
				'scale/500/' + GOOD_IMAGE,
		])

	def test_parse_url__scale_bad(self):
		self.assert_bad_urls([
				'scale/x/' + GOOD_IMAGE,
				'scale/50/a.jpg',
				'scale/500',
		])

	def test_parse_url__width(self):
		self.assert_good_urls([
				'width/100/' + GOOD_IMAGE,
		])

	def test_parse_url__width_bad(self):
		self.assert_bad_urls([
				'width/x/' + GOOD_IMAGE,
				'width/' + GOOD_IMAGE,
		])

	def test_parse_url__height(self):
		self.assert_good_urls([
				'height/100/' + GOOD_IMAGE,
		])

	def test_parse_url__height_bad(self):
		self.assert_bad_urls([
				'height/x/' + GOOD_IMAGE,
				'height/' + GOOD_IMAGE,
		])

	def test_parse_url__fit(self):
		self.assert_good_urls([
				'fit/100x100/' + GOOD_IMAGE,
		])

	def test_parse_url__fit_bad(self):
		self.assert_bad_urls([
				'fit/x/' + GOOD_IMAGE,
				'fit/-1x100/' + GOOD_IMAGE,
				'fit/' + GOOD_IMAGE,
				'fit/48x48/a.jpg',
		])

	def test_parse_url__thumb(self):
		self.assert_good_urls([ 
				'thumb/small/' + GOOD_IMAGE,
				'thumb/medium/' + GOOD_IMAGE,
				'thumb/large/' + GOOD_IMAGE,
				'thumb/50/' + GOOD_IMAGE,
				'thumb/180/' + GOOD_IMAGE,
		])

	def test_parse_url__thumb_bad(self):
		self.assert_bad_urls([
				'thumb/' + GOOD_IMAGE,
				'thumbnail/' + GOOD_IMAGE,
		])

	def test_parse_url__square(self):
		self.assert_good_urls([ 
				'square/180/' + GOOD_IMAGE,
				'square/1/' + GOOD_IMAGE,
		])

	def test_parse_url__square_bad(self):
		self.assert_bad_urls([
				'square/x/' + GOOD_IMAGE,
				'square/-1/' + GOOD_IMAGE,
		])

	def test_parse_url__rotate(self):
		self.assert_good_urls([ 
				'rotate/180/' + GOOD_IMAGE,
				'rotate/-80/' + GOOD_IMAGE,
				'rotate/540/' + GOOD_IMAGE,
		])

	def test_parse_url__crop(self):
		self.assert_good_urls([ 
				'crop/50,50,100x100/' + GOOD_IMAGE,
				'crop/100,100,25x25/' + GOOD_IMAGE,
		])

	def test_parse_url__bw(self):
		self.assert_good_urls([ 
				'bw/' + GOOD_IMAGE,
				'blackwhite/' + GOOD_IMAGE,
		])

	def test_parse_url__invert(self):
		self.assert_good_urls([ 
				'invert/' + GOOD_IMAGE,
		])

	def test_parse_url__blur(self):
		self.assert_good_urls([ 
				'blur/' + GOOD_IMAGE,
		])

	def test_parse_url__sharpen(self):
		self.assert_good_urls([ 
				'sharpen/' + GOOD_IMAGE,
		])

	def test_parse_url__convert(self):
		self.assert_good_urls([ 
				'convert/png/' + GOOD_IMAGE,
				'convert/jpg/' + GOOD_IMAGE,
				'convert/jpeg/' + GOOD_IMAGE,
				'convert/JPEG/' + GOOD_IMAGE,
				'convert/PNG/' + GOOD_IMAGE,
		])

	def test_parse_url__convert_bad(self):
		self.assert_bad_urls([ 
				'convert/pg/' + GOOD_IMAGE,
				'convert/g/' + GOOD_IMAGE,
				'convert/20/' + GOOD_IMAGE,
		])

	def test_parse_url__watermark(self):
		self.assert_good_urls([ 
				'wm/some+stuff/' + GOOD_IMAGE,
				'wm/patrickomatic.com/' + GOOD_IMAGE,
				'watermark/patrickomatic.com/' + GOOD_IMAGE,
		])

	def test_parse_url__combined(self):
		self.assert_good_urls([ 
				'crop/50,50,100x100/bw/resize/50x50/rotate/180/square/50/blur/sharpen/convert/png/' + GOOD_IMAGE,
				'square/500/resize/1000x1000/bw/sharpen/blur/square/60/square/60/' + GOOD_IMAGE,
				])

	def assert_good_urls(self, urls):
		for url in urls:
			self.cr = CommandRunner()	
			self.cr.parse_url(url)
			self.assert_(self.cr.todo)

	def assert_bad_urls(self, urls):
		for url in urls:
			self.cr = CommandRunner()	
			self.cr.parse_url(url)
			self.assertFalse(self.cr.todo)


	def test_run_commands(self):
		self.__mock_s3()

		cr = CommandRunner()
		cr.parse_url('resize/50x50/' + GOOD_IMAGE)
		self.assert_(cr.todo)
		self.assert_(cr.run_commands())

		self.mox.VerifyAll()

	def test_run_commands__thumb_pixel(self):
		self.__mock_s3()

		cr = CommandRunner()
		cr.parse_url('thumb/64/' + GOOD_IMAGE)
		self.assert_(cr.todo)
		self.assert_(cr.run_commands())

		self.mox.VerifyAll()

	def test_run_commands__404(self):
		cr = CommandRunner()
		cr.parse_url('resize/50x50/patrickomatic.com/afdshkfdsa.html')
		self.assert_(cr.todo)

		self.assertRaises(ImageNotFoundException, cr.run_commands)

	def test_run_commands__bad_content_type(self):
		cr = CommandRunner()
		cr.parse_url('resize/50x50/patrickomatic.com/index.html')
		self.assert_(cr.todo)

		self.assertRaises(ImageNotFoundException, cr.run_commands)

	def test_run_commands__image_changed(self):
		image = ModifiedImage.objects.get(id=1)
		image.original_file_hash = 'asfdlj'
		image.last_modified = None
		image.save()

		hash = image.hash

		cr = CommandRunner()
		cr.parse_url(image.get_absolute_url()[1:])
		self.assert_(cr.todo)
		self.assert_(cr.run_commands())

		image = ModifiedImage.objects.get(hash=hash)

	def __mock_s3(self):
		self.mox.StubOutWithMock(boto.s3.key, '__init__')
		boto.s3.key.Key(mox.IgnoreArg())
		self.mox.StubOutWithMock(boto.s3.key.Key, 'set_contents_from_filename')
		boto.s3.key.Key.set_contents_from_filename(mox.IgnoreArg(), mox.IgnoreArg())
		self.mox.ReplayAll()


class ValidatorTest(TestCase):
	def test_validate_percent(self):
		try:
			validate_percent(1)
			validate_percent(2)
			validate_percent(20)
			validate_percent(35)
			validate_percent(90)
			validate_percent(99)
		except ValueError:
			self.fail()
		
	def test_validate_percent__invalid(self):
		self.assertRaises(ValueError, validate_percent, 0)
		self.assertRaises(ValueError, validate_percent, 10000)
		self.assertRaises(ValueError, validate_percent, -50)
		self.assertRaises(ValueError, validate_percent, 1001)


	def test_validate_x_y(self):
		try:
			validate_x_y(1, 1)
			validate_x_y(5, 10)
		except ValueError:
			self.fail()

	def test_validate_x_y__invalid(self):
		self.assertRaises(ValueError, validate_x_y, -1, -1)
		self.assertRaises(ValueError, validate_x_y, 0, -1)
		self.assertRaises(ValueError, validate_x_y, -1, 0)


	def test_validate_width_height(self):
		try:
			validate_width_height(1, 50)
			validate_width_height(20, 1000)
			validate_width_height(1, 5000)
		except ValueError:
			self.fail()

	def test_validate_width_height__invalid(self):
		self.assertRaises(ValueError, validate_width_height, 0, -1)
		self.assertRaises(ValueError, validate_width_height, -1, 20)
		self.assertRaises(ValueError, validate_width_height, 20, 5001)
		self.assertRaises(ValueError, validate_width_height, 5001, 20)


class ImageViewsTest(TestCase):
	def test_modify(self):
		response = self.client.get('/bw/' + GOOD_IMAGE)
		self.assert_(response.status_code == 302)

		response = self.client.get('/square/300/' + GOOD_IMAGE)
		self.assert_(response.status_code == 302)

	def test_modify__unicode_x(self):
		response = self.client.get('/resize/300%C3%97400/' + GOOD_IMAGE)
		self.assert_(response.status_code == 302)

	def test_modify__http(self):
		response = self.client.get('/resize/300%C3%97400/http://' + GOOD_IMAGE)
		self.assert_(response.status_code == 302)

	def test_modify__http_single_slash(self):
		response = self.client.get('/resize/300%C3%97400/http:/' + GOOD_IMAGE)
		self.assert_(response.status_code == 302)

	def test_modify__site_exists(self):
		site = Site(domain_name='patrickomatic.com')
		site.save()

		response = self.client.get('/square/300/' + GOOD_IMAGE)
		self.assert_(response.status_code == 302)

	def test_modify__over_quota(self):
		site = Site(domain_name='patrickomatic.com')
		site.usage = 60 * 1024 * 1024 * 1024
		site.save()

		response = self.client.get('/square/300/' + GOOD_IMAGE)
		self.assertRedirects(response, 'http://urlimg.com/media/over_quota.png')

	def test_modify__suspended(self):
		site = Site(domain_name='patrickomatic.com')
		site.disabled = True
		site.save()

		response = self.client.get('/square/300/' + GOOD_IMAGE)
		self.assertRedirects(response, 'http://urlimg.com/media/suspended.png')

	def test_modify__image_404(self):
		response = self.client.get('/square/300/patrickomatic.com/foo.jpg')
		self.assert_(response.status_code == 404)


class ImageModelsTest(TestCase):
	def test_domain_name(self):
		self.assert_('patrickomatic.com' == domain_name('patrickomatic.com'))
		self.assert_('www.patrickomatic.com' == domain_name('www.patrickomatic.com/'))
		self.assert_('www.patrickomatic.com' == domain_name('http://www.patrickomatic.com/'))
		self.assert_('www.patrickomatic.com' == domain_name('http:/www.PATRICKOMATIC.com/'))
		self.assert_('www.patrickomatic.com' == domain_name('http://www.patrickomatic.com/a bunch of stuff'))
		self.assert_('images.patrickomatic.com' == domain_name('http://images.patrickomatic.com/'))

	def test_readable_bytes(self):
		self.assert_('0 KB' == readable_bytes(1))
		self.assert_('0 KB' == readable_bytes(100))
		self.assert_('0 KB' == readable_bytes(1023))
		self.assert_('1.00 KB' == readable_bytes(1024))
		self.assert_('2.00 KB' == readable_bytes(2048))
		self.assert_('2.50 KB' == readable_bytes(2560))
		self.assert_('1.00 MB' == readable_bytes(1024 * 1024))
		self.assert_('3.00 MB' == readable_bytes(3 * 1024 * 1024))
		self.assert_('1.00 GB' == readable_bytes(1024 * 1024 * 1024))
		self.assert_('3.00 GB' == readable_bytes(3 * 1024 * 1024 * 1024))
		self.assert_('1.00 TB' == readable_bytes(1024 * 1024 * 1024 * 1024))
		self.assert_('3.00 TB' == readable_bytes(3 * 1024 * 1024 * 1024 * 1024))
		self.assert_('1024.00 TB' == readable_bytes(1024 * 1024 * 1024 * 1024 * 1024))
		self.assert_('3072.00 TB' == readable_bytes(3 * 1024 * 1024 * 1024 * 1024 * 1024))

	def test_url_path(self):
		self.assert_('/foo/poo/foo.jpg' == url_path('http://google.com/foo/poo/foo.jpg'))
		self.assert_('/foo/poo/Foo.jpg' == url_path('http://www.google.com/foo/poo/Foo.jpg'))
		self.assert_('/foo/poo/foo.jpg' == url_path('google.com/foo/poo/foo.jpg'))

	def test_sanitize_url(self):
		self.assert_(sanitize_url('patrickomatic.com/foo.jpg') == 'http://patrickomatic.com/foo.jpg')
		self.assert_(sanitize_url('http://patrickomatic.com/foo.jpg') == 'http://patrickomatic.com/foo.jpg')
		self.assert_(sanitize_url('www.patrickomatic.com/foo.jpg') == 'http://www.patrickomatic.com/foo.jpg')
		self.assert_(sanitize_url('http:/www.patrickomatic.com/foo.jpg') == 'http://www.patrickomatic.com/foo.jpg')

	def test_sanitize_url__quote_space(self):
		self.assert_(sanitize_url('patrickomatic.com/foo poo.jpg') == 'http://patrickomatic.com/foo%20poo.jpg')
		self.assert_(sanitize_url('http://patrickomatic.com/foo poo.jpg') == 'http://patrickomatic.com/foo%20poo.jpg')
		self.assert_(sanitize_url('http:/www.patrickomatic.com/foo poo.jpg') == 'http://www.patrickomatic.com/foo%20poo.jpg')

	def test_valid_image_path(self):
		self.assert_(valid_image_path('patrickomatic.com/foo.jpg'))
		self.assert_(valid_image_path('http://patrickomatic.com/foo.jpg'))
		self.assert_(valid_image_path('http://www.patrickomatic.com/foo.jpg'))
		self.assert_(valid_image_path('www.patrickomatic.com/foo.jpg'))
		self.assert_(valid_image_path('http://www.patrickomatic.com/poo/foo.jpg'))
		self.assert_(valid_image_path('http:/www.patrickomatic.com/poo/foo.jpg'))

	def test_valid_image_path__invalid(self):
		self.assert_(not valid_image_path('http://patrickomatic.com/'))
		self.assert_(not valid_image_path('a.jpg'))
		self.assert_(not valid_image_path('foo/a.jpg'))


	def test_retry(self):
		times_called = [0]

		@retry(5, ValueError)
		def retry_tester():
			times_called[0] += 1
			raise ValueError("success")

		self.assertRaises(ValueError, retry_tester)
		self.assert_(times_called[0] == 5)

	def test_retry__eventually_succeed(self):
		times_called = [0]

		@retry(5, Exception)
		def retry_tester():
			times_called[0] += 1
			if times_called[0] < 3:
				raise Exception

			return "success"

		self.assert_("success" == retry_tester())
		self.assert_(times_called[0] == 3)


class ModifiedImageTest(TestCase):
	fixtures = ['image']

	def setUp(self):
		self.image = ModifiedImage.objects.get(id=1)
		self.image.site = Site.objects.get(domain_name='patrickomatic.com')
		self.mox = mox.Mox()

	def tearDown(self):
		self.mox.UnsetStubs()

	
	def test_save(self):
		ModifiedImage(hash='xxx', original_location='/xxx.jpg', operations='resize/50x50', size=500, site=self.image.site).save()

		self.assert_(ModifiedImage.objects.get(hash='xxx'))

	def test_refresh(self):
		last_id = self.image.id
		self.assert_(self.image.refresh())
		self.assert_(self.image.id != last_id)


	def test_delete(self):
		self.mox.StubOutWithMock(boto.s3.key, '__init__')
		boto.s3.key.Key(mox.IgnoreArg())
		self.mox.StubOutWithMock(boto.s3.key.Key, 'delete')
		boto.s3.key.Key.delete()
		self.mox.StubOutWithMock(boto.s3.key.Key, 'close')
		boto.s3.key.Key.close()
		self.mox.ReplayAll()

		id = self.image.id
		self.image.delete()

		self.assertRaises(ModifiedImage.DoesNotExist, ModifiedImage.objects.get, id=id)
				
		self.mox.VerifyAll()


	def test_get_thumbnail_url(self):
		self.assert_(re.match(r'/%sinternalthumb/\w+/%s%s' \
				% (self.image.operations, self.image.site.domain_name, self.image.original_location),
				self.image.get_thumbnail_url()))


	def test_get_account_url(self):
		self.assertEquals('/account/sites/patrickomatic.com/images/1/', self.image.get_account_url())


	def test_get_absolute_url(self):
		self.assertEquals('/' + self.image.operations + self.image.site.domain_name \
				+ self.image.original_location, self.image.get_absolute_url())

class ImageModuleTest(TestCase):
	def test_color_to_rgb(self):
		self.assert_(color_to_rgb('red') == (255, 0, 0))
		self.assert_(color_to_rgb('green') == (0, 128, 0))
		self.assert_(color_to_rgb('foo') == None)

	def test_hex_to_decimal(self):
		self.assert_(hex_to_decimal('FF') == 255)
		self.assert_(hex_to_decimal('0') == 0)
		self.assert_(hex_to_decimal('CC') == 204)

	
	def test_hex_to_rgb(self):
		self.assert_(hex_to_rgb('#FF00CC') == (255, 0, 204))
		self.assert_(hex_to_rgb('000080') == (0, 0, 128))
		self.assert_(hex_to_rgb('0CF') == (0, 204, 255))
		self.assert_(hex_to_rgb('#0CF') == (0, 204, 255))
		self.assert_(hex_to_rgb('#fooooooo') == None)



import sys, unicodedata
from django.http import Http404, HttpResponseRedirect
from django.conf import settings

from urlimaging.models import *


LATIN_ASCII_MAP = { 
	0xc0: 'A', 0xc1: 'A', 0xc2: 'A', 0xc3: 'A', 0xc4: 'A', 0xc5: 'A', 0xc6: 'Ae', 0xc7: 'C', 
	0xc8: 'E', 0xc9: 'E', 0xca: 'E', 0xcb: 'E', 0xcc: 'I', 0xcd: 'I', 0xce: 'I', 0xcf: 'I', 
	0xd0: 'Th', 0xd1: 'N', 0xd2: 'O', 0xd3: 'O', 0xd4: 'O', 0xd5: 'O', 0xd6: 'O', 0xd8: 'O', 
	0xd9: 'U', 0xda: 'U', 0xdb: 'U', 0xdc: 'U', 0xdd: 'Y', 0xde: 'th', 0xdf: 'ss', 0xe0: 'a', 
	0xe1: 'a', 0xe2: 'a', 0xe3: 'a', 0xe4: 'a', 0xe5: 'a', 0xe6: 'ae', 0xe7: 'c', 0xe8: 'e', 
	0xe9: 'e', 0xea: 'e', 0xeb: 'e', 0xec: 'i', 0xed: 'i', 0xee: 'i', 0xef: 'i', 0xf0: 'th', 
	0xf1: 'n', 0xf2: 'o', 0xf3: 'o', 0xf4: 'o', 0xf5: 'o', 0xf6: 'o', 0xf8: 'o', 0xf9: 'u', 
	0xfa: 'u', 0xfb: 'u', 0xfc: 'u', 0xfd: 'y', 0xfe: 'th', 0xff: 'y', 0xa1: '!', 
	0xa2: '{cent}', 0xa3: '{pound}', 0xa4: '{currency}', 0xa5: '{yen}', 0xa6: '|', 
	0xa7: '{section}', 0xa8: '{umlaut}', 0xa9: '{C}', 0xaa: '{^a}', 0xab: '<<', 0xac: '{not}', 
	0xad: '-', 0xae: '{R}', 0xaf: '_', 0xb0: '{degrees}', 0xb1: '{+/-}', 0xb2: '{^2}', 
	0xb3: '{^3}', 0xb4: "'", 0xb5: '{micro}', 0xb6: '{paragraph}', 0xb7: '*', 
	0xb8: '{cedilla}', 0xb9: '{^1}', 0xba: '{^o}', 0xbb: '>>', 0xbc: '{1/4}', 0xbd: '{1/2}', 
	0xbe: '{3/4}', 0xbf: '?', 0xd7: '*', 0xf7: '/', 0x00d7: 'x', 
}


def latin1_to_ascii(uni):
	r = ''
	for i in uni:
		if LATIN_ASCII_MAP.has_key(ord(i)):
			r += LATIN_ASCII_MAP[ord(i)]
		elif ord(i) >= 0x80:
			pass
		else:
			r += str(i)
	return r


def modify(request, url):
	if request.META['QUERY_STRING']:
		url = url + "?" + request.META['QUERY_STRING']

	cr = CommandRunner(latin1_to_ascii(url))

	if not settings.IMAGE_WHITELIST_FN(cr.url):
		raise Http404	

	try:
		if not cr.todo: 
			print >>sys.stderr, "Unable to parse url: %s" % url
			raise Http404
		
		img = cr.run_commands()
		if not img: 
			print >>sys.stderr, "Unable to run commands: %s" % url
			raise Http404

		return HttpResponseRedirect(img)
	except ImageNotFoundException:
		raise Http404
	
	raise Http404



import urllib, urllib2, os, hashlib, re, datetime

from django.conf import settings
from django.contrib.auth.models import User
from django.db import models

from urlimaging.image import *


def domain_name(url):
	return re.sub(r'/.*$', '', re.sub(r'^(http://?)?', '', url)).lower()


def url_path(url):
	return re.sub(r'^[^/]+', '', re.sub(r'^http://?', '', url))


class Site(models.Model):
	domain_name = models.CharField(max_length=255, unique=True, db_index=True) 

	def get_absolute_url(self):
		return self.domain_name

	def __unicode__(self):
		return self.domain_name


class ModifiedImage(models.Model):
	hash = models.CharField(max_length=56, unique=True, db_index=True) 
	ext = models.CharField(max_length=10, db_index=True)
	operations = models.CharField(max_length=4096) 
	last_modified = models.CharField(max_length=255, null=True)
	last_checked = models.DateTimeField(db_index=True, default=datetime.datetime.now())
	original_location = models.CharField(max_length=4096)
	original_file_hash = models.CharField(max_length=56)
	site = models.ForeignKey(Site)
	size = models.PositiveIntegerField(default=0)


	def delete(self):
		settings.IMAGE_STORAGE.delete_image(self.hash)

		models.Model.delete(self)

	def refresh(self):
		cr = CommandRunner(self.get_absolute_url()[1:])
		if cr.todo:
			self.delete()
			if cr.run_commands():
				return True

		return False

	def get_absolute_url(self):
		return '/' + self.operations + self.site.domain_name + self.original_location

	def __unicode__(self):
		return self.get_absolute_url()


# XXX use a setting for the directory
def file_location(hash, ext):
	return '/tmp/' + hash + ext


def sanitize_url(url):
	domain, url = domain_name(url), url_path(url)
	return 'http://' + domain + urllib.quote(url, safe="%/:=&?~#+!$,;'@()*[]")


def valid_image_path(url):
	return re.match(r'^(http://?)?[\w\-\.]+\.\w+/.+$', url, re.I)


class ImageNotFoundException(Exception):
	pass


class CommandRunner:
	def __init__(self, url=None):
		self.todo = []
		self.url = ""
		self.filename = ""
		self.hashed_filename = ""
		self.ext = ""
		self.hash = ""
		self.operations = ""

		if url: self.parse_url(url)


	def parse_url(self, url):
		did_op = True

		# apply the operations to the url until none are left
		while did_op:
			did_op = False
			for op in COMMANDS:
				m = re.match(op['regex'], url)
				if m:
					self.operations += m.group().lower()
					self.todo.append((op['fn'], m.groups()))
					url = re.sub(op['regex'], '', url)
					did_op = True
					break


		if self.todo and valid_image_path(url):
			self.url = sanitize_url(url)
			self.ext = os.path.splitext(re.sub(r'\?.+$', '', url))[-1]
			if not self.ext:
				self.ext = '.jpg'

			self.hash = hashlib.sha224(self.operations + url).hexdigest()
			self.filename = file_location(self.hash, self.ext)
			self.hashed_filename = "%s%s" % (self.hash, self.ext)
		else:
			self.todo = []


	def get_image(self, image):
		""" Get the image and return true or false if it's changed or not.  It can tell
		    if it's been used by either using the last-modified header saved from a previous
		    request, or if that doesn't exist, it compares a hash of the file's contents """
		image.last_checked = datetime.datetime.now()
		# if it exists, always save here that it's been checked
		if image.id: image.save()

		request = urllib2.Request(self.url)

		if image.last_modified:
			request.add_header('If-Modified-Since', image.last_modified)

		try:
			remote = urllib2.build_opener().open(request)
		except urllib2.HTTPError as e:
			if e.code == 404: raise ImageNotFoundException()
			# 304 Not Modified
			return False
		except urllib2.URLError as e:
			# requesting host down?
			return False

		content_type = remote.info().getheader('content-type').lower()
		if not content_type.startswith('image/') \
				and not content_type.endswith("/octet-stream"):
			remote.close()
			raise ImageNotFoundException()

		hasher = hashlib.sha224()
		local = open(self.filename, 'w')

		# read file in 4K chunks and make a hash of it
		size = 0
		while True:
			chunk = remote.read(4096)
			if not chunk: break

			size += len(chunk)

			# don't process images larger than 10 MBS
			if size > TEN_MBS: return False

			local.write(chunk)
			hasher.update(chunk)

		local.close()
		remote.close()

		remote_file_hash = hasher.hexdigest()
		if image.original_file_hash == remote_file_hash:
			# compared the hashes and they are still the same (no change)
			os.unlink(self.filename)
			return False

		# if it's changed, delete the previous
		if image.id:
			image.delete()

		image.original_file_hash = remote_file_hash
		image.last_modified = remote.info().getheader('last-modified')

		return True


	def run_commands(self, user=None):
		check_remote_image = False
		try:
			image = ModifiedImage.objects.get(hash=self.hash)	
		except ModifiedImage.DoesNotExist:
			# need to create the image
			image = ModifiedImage(hash=self.hash, \
					original_location=url_path(self.url), \
					operations=self.operations, ext=self.ext)

			# do we need to create the domain too?
			try:
				site = Site.objects.get(domain_name=domain_name(self.url))

			except Site.DoesNotExist:
				site = Site(domain_name=domain_name(self.url))
				site.save()

			image.site = site
			check_remote_image = True
		else:
			two_hours_ago = datetime.datetime.now() - datetime.timedelta(hours=2)
			check_remote_image = image.last_checked < two_hours_ago


		if check_remote_image and self.get_image(image):
			# apply all of the transformations
			for fn, group in self.todo:
				try: 
					fn(self.filename, *group)
				except ValueError as e:
					break

			settings.IMAGE_STORAGE.save_image(self.hashed_filename, self.filename)
			image.size = os.path.getsize(self.filename)

			os.unlink(self.filename)

		if check_remote_image:
			image.save()

		return settings.IMAGE_STORAGE.get_image_url(self.hashed_filename)



import time, shutil, os

from django.conf import settings


class ImageStorage:
	def delete_image(self, hash):
		raise Exception('delete_image not implemented')

	def save_image(self, filename):
		raise Exception('save_image not implemented')

	def get_image_url(self, hash):
		raise Exception('get_image_url not implemented')

	def get_required_settings(self):
		raise Exception('get_required_settings not implemented')


def retry(times, ex):
	""" A decorator which can be called as:

		@retry(5, S3DataError)
		def fn():
			pass

	    and will call fn and if it throws an exception, it will keep
	    trying 5 times. """
	def retry_wrap(fn):
		def fn_wrap(*args, **kwargs):
			for i in range(times-1):
				try:
					return fn(*args, **kwargs)
				except ex:
					pass
			return fn(*args, **kwargs)
		return fn_wrap
	return retry_wrap


class S3ImageStorage(ImageStorage):
	def __init__(self):
		from boto.s3.connection import S3Connection

		self.connection = S3Connection(settings.AWS_ACCESS_KEY_ID, settings.AWS_SECRET_ACCESS_KEY)
		self.bucket = self.connection.get_bucket(settings.S3_BUCKET_NAME)


	def delete_image(self, hash):
		from boto.s3.key import Key

		k = Key(self.bucket, hash)
		k.delete()
		k.close()


	@retry(3, Exception)
	def save_image(self, hash, filename):
		from boto.s3.key import Key

		key = Key(self.bucket, hash)
		key.set_contents_from_filename(filename, 
				{'Cache-Control': 'public, max-age=%d' % settings.S3_EXPIRES,
				'Expires': time.asctime(time.gmtime(time.time() + settings.S3_EXPIRES)) })
		key.close()


	def get_image_url(self, hash):
		from boto.s3.key import Key

		key = Key(self.bucket, hash)

		ret = key.generate_url(settings.S3_EXPIRES, 'GET', 
				{'Cache-Control': 'public, max-age=%d' % settings.S3_EXPIRES,
				'Expires': time.asctime(time.gmtime(time.time() + settings.S3_EXPIRES)) })
		key.close()

		return ret


	def get_required_settings(self):
		return ['AWS_ACCESS_KEY_ID', 'AWS_SECRET_ACCESS_KEY', 'S3_BUCKET_NAME']


class LocalImageStorage(ImageStorage):
	def get_storage_dir(self):
		try:
			return getattr(settings, 'IMAGE_STORAGE_DIR')
		except AttributeError:
			return settings.MEDIA_ROOT


	def delete_image(self, hash):
		os.unlink(os.path.join(self.get_storage_dir(), hash))


	def save_image(self, hash, filename):
		shutil.copyfile(filename, os.path.join(self.get_storage_dir(), hash))


	def get_image_url(self, hash):
		# XXX check for a setting
		return settings.MEDIA_URL + "/" + hash


	def get_required_settings(self):
		return []


class SCPImageStorage(ImageStorage):
	def identity_file_str(self):
		try:
			return "-i " + settings.SSH_IDENTITY_FILE
		except AttributeError:
			return ""

	def delete_image(self, hash):
		os.system("ssh %(identify_file)s %(ssh_user)s \"rm %(file_path)s\"" % {
					'identity_file': self.identity_file_str(),
					'ssh_user': settings.SSH_MEDIA_USER,
					'file_path': os.path.join(settings.SSH_MEDIA_PATH, hash)})

	def save_image(self, filename):
		os.system("scp %(identity_file)s %(filename)s %(ssh_user)s:%(file_path)s" % {
					'identity_file': self.identity_file_str(),
					'ssh_user': settings.SSH_MEDIA_USER, 
					'filename': filename,
					'file_path': os.path.join(settings.SSH_MEDIA_PATH, hash)})

	def get_image_url(self, hash):
		url = settings.PROCESSED_MEDIA_URL
		if url.endswith("/"):
			return url + hash
		else:
			return url + '/' + hash

	def get_required_settings(self):
		return ['PROCESSED_MEDIA_URL', 'SSH_MEDIA_USER', 'SSH_MEDIA_PATH',]




def validate_percent(percent):
	if percent < 1 or percent > 1000:
		raise ValueError('Must be between 0 and 1000: ' + str(percent))

def validate_x_y(x, y):
	if x < 0 or y < 0:
		raise ValueError('Must be positive: x=' + str(x) + ',y=' + str(y))

def validate_width_height(width, height):
	if width < 1 or height < 1 or width > 5000 or height > 5000:
		raise ValueError('Must be between 0 and 5,000: width=' + str(width) + ',height=' + str(height))



"""

Django URL-imaging

Author: Patrick Carroll <patrick@patrickomatic.com>
Version: 0.1

"""
from django.core.exceptions import ImproperlyConfigured
from django.conf import settings

from urlimaging.backends.default import *

VERSION = 0.1

# set up the storage backend - default to S3
try:
	settings.IMAGE_STORAGE = eval('%s()' % getattr(settings, 'IMAGE_STORAGE_BACKEND'))
except AttributeError:
	settings.IMAGE_STORAGE = S3ImageStorage()

# default S3_EXPIRES
try:
	getattr(settings, 'S3_EXPIRES')
except AttributeError:
	settings.S3_EXPIRES = 2 * 24 * 60 * 60

# depending on the backend used, make sure that all required settings are supplied
for setting in settings.IMAGE_STORAGE.get_required_settings():
	try:
		getattr(settings, setting)
	except AttributeError:
		raise ImproperlyConfigured("You must set %s in your settings.py" % setting)
	

# the function which decides whether or not to process the image
try:
	getattr(settings, 'IMAGE_WHITELIST_FN')
except AttributeError:
	# XXX should this default to True?
	settings.IMAGE_WHITELIST_FN = lambda url: settings.MEDIA_URL in url


# the font settings for watermarking
try:
	getattr(settings, 'FONT_PATH')
except AttributeError:
	settings.FONT_PATH = '/usr/share/fonts/truetype/freefont/FreeSansBold.ttf'


# days before images that haven't been visited are considered for deletion
try:
	getattr(settings, 'IMAGE_EXPIRATION_DAYS')
	# XXX verify that it's an int
except AttributeError:
	settings.IMAGE_EXPIRATION_DAYS = 7



import re, os
from math import atan, degrees
from PIL import Image, ImageChops, ImageFilter, ImageDraw, ImageFont

from django.conf import settings

from urlimaging.validator import *


TEN_MBS = 10 * 1024 * 1024 * 1024
RGB_ONLY_FORMATS = set(['.jpg', '.jpeg'])


def with_image(fn):
	""" Decorator that handles the open and closing of an image.  """
	def process(*args):
		img = fn(Image.open(args[0]), *args[1:])

		ext = os.path.splitext(args[0])[1].lower()
		if ext in RGB_ONLY_FORMATS and img.mode != 'RGB':
			img = img.convert('RGB')
		elif ext == '.png' and img.mode in ('P', 'I', 'RGB'):
			img = img.convert('RGBA').convert('P', palette=Image.ADAPTIVE)

		img.save(args[0], quality=95, **img.info)

	return process


@with_image
def resize(img, width, height):
	width, height = int(width), int(height)
	validate_width_height(width, height)
	return img.resize((int(width), int(height)), Image.ANTIALIAS)


@with_image
def scale(img, percent):
	percent = int(percent)
	validate_percent(percent)

	p = percent * 0.01

	w, h = img.size
	return img.resize((int(w * p), int(h * p)), Image.ANTIALIAS)


@with_image
def fit(img, width, height):
	img.thumbnail((int(width), int(height)), Image.ANTIALIAS)
	return img


@with_image
def zoom(img, width, height):
	w, h = img.size

	if w > h:
		w = int(width)
		h = int(float(h) * (float(width) / float(w)))
		img = img.resize((w, h), Image.ANTIALIAS)
		to_crop = int((h - height) / 2.0)

		return img.crop((0, to_crop, width, height + to_crop))
	else:
		w = int(float(w) * (float(height) / float(h)))
		h = int(height)
		img = img.resize((w, h), Image.ANTIALIAS) 
		to_crop = int((w - width) / 2.0)

		return img.crop((to_crop, 0, width + to_crop, height))


@with_image
def width(img, width):
	w, h = img.size
	return img.resize((int(width), int(float(h) * (float(width) / float(w)))), Image.ANTIALIAS)


@with_image
def height(img, height):
	w, h = img.size
	return img.resize((int(float(w) * (float(height) / float(h))), int(height)), Image.ANTIALIAS)


@with_image
def square(img, size):
	size = int(size)
	validate_width_height(size, size)

	width, height = img.size

	half_size = size / 2.0
	# already square?
	if width == height:
		img = img.resize((size, size), Image.ANTIALIAS)
		return img
	elif width > height:
		# resize proportionally to get height == size
		resize_amount = float(size) / float(height)
		width = int(width * resize_amount)
		img = img.resize((width, size), Image.ANTIALIAS)

		return img.crop((int(width / 2.0 - half_size), 0, int(width / 2.0 + half_size), size))
	else:
		# get width == size
		resize_amount = float(size) / float(width)
		height = int(height * resize_amount)
		img = img.resize((size, height), Image.ANTIALIAS)

		return img.crop((0, int(height / 2.0 - half_size), size, int(height / 2.0 + half_size)))


@with_image
def crop(img, x, y, width, height):
	x, y, width, height = int(x), int(y), int(width), int(height)
	validate_x_y(x, y)
	validate_width_height(width, height)

	return img.crop((x, y, x + width, y + height))


@with_image
def rotate(img, degrees):
	return img.rotate(-int(degrees))


@with_image
def thumbnail(img, size):
	THUMBNAIL_SIZES = { 
		'small': 64,
		'medium': 96,
		'large': 128,
	}

	if size.isdigit():
		s = int(size)
	elif size in THUMBNAIL_SIZES:
		s = THUMBNAIL_SIZES[size]
	else:
		s = THUMBNAIL_SIZES['medium']

	img.thumbnail((s, s), Image.ANTIALIAS)
	return img


@with_image
def black_and_white(img):
	return img.convert('L')


@with_image
def invert(img):
	return ImageChops.invert(img)


@with_image
def blur(img):
	return img.filter(ImageFilter.BLUR)

@with_image
def sharpen(img):
	return img.filter(ImageFilter.SHARPEN)


def convert(filename, format):
	img = Image.open(filename)
	if img.mode != 'RGB':
		img = img.convert('RGB')

	# for some reason this chokes on 'JPG'
	format = format.upper()
	if format == 'JPG':
		format = 'JPEG'

	img.save(filename + '.' + format, format, quality=95)


@with_image
def watermark(img, text):
	if img.mode != 'RGB':
		img = img.convert('RGB')

	watermark = Image.new('RGBA', (img.size[0], img.size[1]))
	draw = ImageDraw.ImageDraw(watermark, 'RGBA')

	size = 0
	while True:
		size += 1
		nextFont = ImageFont.truetype(settings.FONT_PATH, size)
		nextTextWidth, nextTextHeight = nextFont.getsize(text)
		if nextTextWidth + nextTextHeight / 3 > watermark.size[0]:
			break
		font = nextFont
		textWidth, textHeight = nextTextWidth, nextTextHeight
	
	draw.text(((watermark.size[0] - textWidth) / 2,
			(watermark.size[1] - textHeight) / 2), text, font=font, fill=(255,255,255))

	watermark = watermark.rotate(degrees(atan(float(img.size[1]) /
					img.size[0])), Image.BICUBIC)
	mask = watermark.convert('L').point(lambda x: min(x, 55))
	watermark.putalpha(mask)
	img.paste(watermark, None, watermark)

	return img


HTML_COLORS = { 
	'aqua': (0, 255, 255),
	'cyan': (0, 255, 255),
	'black': (0, 0, 0),
	'blue': (0, 0, 255),
	'fuchsia': (255, 0, 255),
	'magenta': (255, 0, 255),
	'gray': (128, 128, 128),
	'grey': (128, 128, 128),
	'green': (0, 128, 0),
	'lime': (0, 255, 0),
	'maroon': (128, 0, 0),
	'navy': (0, 0, 128),
	'olive': (128, 128, 0),
	'purple': (128, 0, 128),
	'red': (255, 0, 0),
	'silver': (192, 192, 192),
	'teal': (0, 128, 128),
	'white': (255, 255, 255),
	'yellow': (255, 255, 0),
}

def color_to_rgb(color):
	try:
		return HTML_COLORS[color.lower()]
	except KeyError:
		return None

def hex_to_decimal(h):
	return int(h, 16)

def hex_to_rgb(hex):
	""" Convert a hexadecimal triplet to an RGB tuple """

	if hex.startswith("#"):
		hex = hex[1:]

	try:
		if len(hex) == 3:
			hex_triplet = map(lambda x: "%s%s" % (x, x), tuple(hex))
			return tuple(map(hex_to_decimal, hex_triplet))
		elif len(hex) == 6:
			return tuple(map(hex_to_decimal, (hex[0:2], hex[2:4], hex[4:6])))
	except ValueError:
		pass

	return None


@with_image
def background(img, color):
	rgb = hex_to_rgb(color)
	if not rgb: rgb = color_to_rgb(color)
	if not rgb: return None

	img = img.convert('RGBA')
	overlayed = Image.new('RGBA', img.size)
	overlayed.paste(rgb)
	overlayed.paste(img, mask=img)

	return overlayed


COMMANDS = [{ 
		'regex': re.compile(r'^resize/(\d+)x(\d+)/', re.I),
		'fn': resize, 
		'title': 'Resize',
		'anchor_name': 'resize',
		'description': 'Resize an image.',
		'format': '<tt>/resize/WIDTHxHEIGHT/IMAGEURL</tt>',
		'arguments': [ 
			('WIDTHxHEIGHT', 'The width and height of the desired image. Expressed as a width and height specification of the form WIDTHxHEIGHT where WIDTH and HEIGHT are both positive numbers.') 
		],
	}, {
		'regex': re.compile(r'^scale/(\d+)/', re.I),
		'fn': scale, 
		'title': 'Scale',
		'anchor_name': 'scale',
		'description': 'Scale an image proportionally by a given percentage',
		'format': '<tt>/scale/PERCENT/IMAGEURL</tt>',
		'arguments': [ 
			('PERCENT', 'A value between 0 and 1000 by which to scale the image.  If greater than 100, the image will be increased in size by that percentage amount.') 
		],
	}, {
		'regex': re.compile(r'^width/(\d+)/', re.I),
		'fn': width, 
		'title': 'Width',
		'anchor_name': 'width',
		'description': 'Scale an image proportionaly to the given width',
		'format': '<tt>/width/SIZE/IMAGEURL</tt>',
		'arguments': [ ('SIZE', 'A positive number specifying the target width of the image.') ],
	}, {
		'regex': re.compile(r'^height/(\d+)/', re.I),
		'fn': height, 
		'title': 'Height',
		'anchor_name': 'height',
		'description': 'Scale an image proportionaly to the given height',
		'format': '<tt>/height/SIZE/IMAGEURL</tt>',
		'arguments': [ 
			('SIZE', 'A positive number specifying the target height of the image.') 
		],
	}, {
		'regex': re.compile(r'^fit/(\d+)x(\d+)/', re.I),
		'fn': fit, 
		'title': 'Fit',
		'anchor_name': 'fit',
		'description': 'Fit an image proportionally to be within a given size',
		'format': '<tt>/fit/WIDTHxHEIGHT/IMAGEURL</tt>',
		'arguments': [ 
			('WIDTHxHEIGHT', 'The box that the image must be resized to fit into. Expressed as a width and height specification of the form WIDTHxHEIGHT where WIDTH and HEIGHT are both positive numbers.') 
		],
	}, {
		'regex': re.compile(r'^zoom/(\d+)x(\d+)/', re.I),
		'fn': zoom, 
		'title': 'Zoom',
		'anchor_name': 'zoom',
		'description': 'Similar to fit, but will zoom in on the image to get one side matching as best as possible.  If the image is wider than tall, it will do the equivalent of /height/HEIGHT/ otherwise it will do the equivalent of /width/WIDTH/',
		'format': '<tt>/zoom/WIDTHxHEIGHT/IMAGEURL</tt>',
		'arguments': [ 
			('WIDTHxHEIGHT', 'The box that the image must be zoomed to fit into. Expressed as a width and height specification of the form WIDTHxHEIGHT where WIDTH and HEIGHT are both positive numbers.') 
		],
	}, {
		'regex': re.compile(r'^thumb(?:nail)?/(small|medium|large|\d+)/', re.I),
		'fn': thumbnail, 
		'title': 'Thumbnail',
		'anchor_name': 'thumbnail',
		'description': 'Create a thumbnail of an image',
		'format': """<tt>/thumbnail/SIZE/IMAGEURL</tt>
				or a shorter version:
			     <tt>/thumb/SIZE/IMAGEURL</tt>""",
		'arguments': [ ('SIZE', 'A number of pixels or a value of either "small" (64 pixels), "medium" (96 pixels) or "large" (128).') ],
	}, {
		'regex': re.compile(r'^square/(\d+)/', re.I),
		'fn': square, 
		'title': 'Square',
		'anchor_name': 'square',
		'description': 'Create a squared version of an image.  This will resize and crop an image to create a squared version with the height and width given.',
		'format': '<tt>/square/SIZE/IMAGEURL</tt>',
		'arguments': [ 
			('SIZE', 'The height and width of the desired square'), 
		],
	}, {
		'regex': re.compile(r'^rotate/(-?\d+)/', re.I),
		'fn': rotate, 
		'title': 'Rotate',
		'anchor_name': 'rotate',
		'description': 'Rotate an image',
		'format': '<tt>/rotate/DEGREES/IMAGEURL</tt>',
		'arguments': [ 
			('DEGREES', 'A value specifying the number of degrees that the image is to be rotated in a clockwise direction.  A negative number will rotate the image in a counter-clockwise direction.') 
		],
	}, {
		'regex': re.compile(r'^crop/(\d+),(\d+),(\d+)x(\d+)/', re.I),
		'fn': crop, 
		'title': 'Crop',
		'anchor_name': 'crop',
		'description': 'Select a rectangular region of an image and remove everything outside of that region.',
		'format': '<tt>/crop/CORNER-X,CORNER-Y,WIDTHxHEIGHT/IMAGEURL</tt>',
		'arguments': [ 
			('CORNER-X', 'The x coordinate of the top left corner of the rectangle to be cropped. Assumes that the top left corner is (0, 0).'),
			('CORNER-Y', 'The y coordinate of the top left corner of the rectangle to be cropped.'),
			('WIDTH', 'The width of the rectangle to be cropped.'),
			('HEIGHT', 'The height of the rectangle to be cropped.'),
		],
	}, {
		'regex': re.compile(r'^(?:watermark|wm)/([^/]+)/', re.I),
		'fn': watermark, 
		'title': 'Watermark',
		'anchor_name': 'watermark',
		'description': 'Overlay some text on the image',
		'format': '<tt>/wm/TEXT/IMAGEURL</tt> or <tt>/watermark/TEXT/IMAGEURL</tt>',
		'arguments': [ 
			('TEXT', 'The text of the watermark.'),
		],
		'note': 'It is recommended that after generating the watermarked image, you remove the original image.  Otherwise it will continue to be publicly available.'
	}, {
		'regex': re.compile(r'^(?:blackwhite|bw)/', re.I),
		'fn': black_and_white, 
		'title': 'Black & White',
		'anchor_name': 'blackwhite',
		'description': 'Convert the image to black and white',
		'format': '<tt>/bw/IMAGEURL</tt> or <tt>/blackwhite/IMAGEURL</tt>',
		'arguments': [ ],
	}, {
		'regex': re.compile(r'^invert/', re.I),
		'fn': invert, 
		'title': 'Invert',
		'anchor_name': 'invert',
		'description': 'Invert the color profile of the image',
		'format': '<tt>/invert/IMAGEURL</tt>',
		'arguments': [ ],
	}, {
		'regex': re.compile(r'^blur/', re.I),
		'fn': blur, 
		'title': 'Blur',
		'anchor_name': 'blur',
		'description': 'Apply a blurring filter to the image',
		'format': '<tt>/blur/IMAGEURL</tt>',
		'arguments': [ ],
	}, {
		'regex': re.compile(r'^sharpen/', re.I),
		'fn': sharpen, 
		'title': 'Sharpen',
		'anchor_name': 'sharpen',
		'description': 'Apply a sharpening filter to the image',
		'format': '<tt>/sharpen/IMAGEURL</tt>',
		'arguments': [ ],
	}, {
		'regex': re.compile(r'^convert/\.?(bmp|gif|im|jpe?g|msp|pcx|pdf|png|ppm|tiff|xbm)/', re.I),
		'fn': convert, 
		'title': 'Convert',
		'anchor_name': 'convert',
		'description': 'Convert the image to a different format',
		'format': '<tt>/convert/FORMAT/IMAGEURL</tt>',
		'arguments': [ 
			('FORMAT', 'The desired image format of the created image.  Allowable values are: bmp, gif, im, jpeg/jpg, msp, pcx, pdf, png, ppm, tiff, xbm')
		],
	}, {
		'regex': re.compile(r'^background/(\w+)/', re.I),
		'fn': background, 
		'title': 'Background',
		'anchor_name': 'background',
		'description': 'Set a custom background color.  The image must support transparency.',
		'format': '<tt>http://urlimg.com/background/FORMAT/IMAGEURL</tt>',
		'arguments': [ 
			('FORMAT', 'Either a color (red, green, blue, etc..) or a hexadecimal triplet (#FF00FF).')
		],
	}, 
]

